# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import numpy as np

from .convert import GRIB_TO_CF, GRIB_TO_XARRAY_PL, GRIB_TO_XARRAY_SFC

LOG = logging.getLogger(__name__)


def save_output_xarray(
    *,
    output,
    target_variables,
    write,
    all_fields,
    ordering,
    lead_time,
    hour_steps,
    lagged,
):
    LOG.info("Converting output xarray to GRIB and saving")

    output["total_precipitation_6hr"] = output.data_vars[
        "total_precipitation_6hr"
    ].cumsum(dim="time")

    all_fields = all_fields.order_by(
        valid_datetime="descending",
        param_level=ordering,
        remapping={"param_level": "{param}{levelist}"},
    )

    for time in range(lead_time // hour_steps):
        for fs in all_fields[: len(all_fields) // len(lagged)]:
            param, level = fs["shortName"], fs["level"]

            if level != 0:
                param = GRIB_TO_XARRAY_PL.get(param, param)
                if param not in target_variables:
                    continue
                values = output.isel(time=time).sel(level=level).data_vars[param].values
            else:
                param = GRIB_TO_CF.get(param, param)
                param = GRIB_TO_XARRAY_SFC.get(param, param)
                if param not in target_variables:
                    continue
                values = output.isel(time=time).data_vars[param].values

            # We want to field north=>south

            values = np.flipud(values.reshape(fs.shape))

            write(
                values,
                template=fs,
                step=(time + 1) * hour_steps,
            )



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging

import climetlab as cml
import numpy as np
import xarray as xr

from .convert import GRIB_TO_XARRAY_PL, GRIB_TO_XARRAY_SFC

LOG = logging.getLogger(__name__)


def forcing_variables_numpy(sample, forcing_variables, dates):
    """Generate variables from climetlabs

    Args:
        date (datetime): Datetime of current time step in forecast
        params (List[str]): Parameters to calculate as constants

    Returns:
        torch.Tensor: Tensor with constants
    """
    ds = cml.load_source(
        "constants",
        sample,
        date=dates,
        param=forcing_variables,
    )

    return (
        ds.order_by(param=forcing_variables, valid_datetime="ascending")
        .to_numpy()
        .reshape(len(forcing_variables), len(dates), 721, 1440)
    )


def create_training_xarray(
    *,
    fields_sfc,
    fields_pl,
    lagged,
    start_date,
    hour_steps,
    lead_time,
    forcing_variables,
    constants,
):
    LOG.info("Creating training dataset")
    # Create Input dataset

    sfc_fields_xarray = (
        fields_sfc.to_xarray().rename(GRIB_TO_XARRAY_SFC).isel(number=0, surface=0)
    )

    pl_fields_xarray = fields_pl.to_xarray().rename(GRIB_TO_XARRAY_PL).isel(number=0)

    sfc_fields_xarray.coords["time"] = [
        datetime.timedelta(hours=hour) for hour in lagged
    ]

    pl_fields_xarray.coords["time"] = [
        datetime.timedelta(hours=hour) for hour in lagged
    ]

    # Combine lagged and future timedeltas
    time_deltas = [
        datetime.timedelta(hours=h)
        for h in lagged
        + [hour for hour in range(hour_steps, lead_time + hour_steps, hour_steps)]
    ]
    datetimes = [start_date() + time_delta for time_delta in time_deltas]
    forcing_numpy = forcing_variables_numpy(fields_sfc, forcing_variables, datetimes)
    # Create an empty training dataset that has all the variables from sfc_fields
    # and pl_fields but nans over the dimensions
    # batch, time, lat, lon, level
    # This is so we can merge the forcing dataset with the training dataset
    # and then drop the batch dimension
    empty_dataset = xr.Dataset(
        {
            "toa_incident_solar_radiation": (
                ["batch", "time", "lat", "lon"],
                forcing_numpy[0:1, :, :, :],
            ),
            "geopotential_at_surface": (
                ["lat", "lon"],
                np.squeeze(
                    sfc_fields_xarray["geopotential_at_surface"].values[0, 0, :, :]
                ),
            ),
            "land_sea_mask": (
                ["lat", "lon"],
                np.squeeze(sfc_fields_xarray["land_sea_mask"].values[0, 0, :, :]),
            ),
        },
        coords={
            "batch": sfc_fields_xarray.coords["batch"],
            "time": time_deltas,
            "lat": sfc_fields_xarray.coords["lat"],
            "lon": sfc_fields_xarray.coords["lon"],
            "level": pl_fields_xarray.coords["level"],
        },
    )
    sfc_fields_xarray = sfc_fields_xarray.drop_vars(
        [
            "geopotential_at_surface",
            "land_sea_mask",
        ]
    )

    # Create a training dataset with all the variables from sfc_fields and pl_fields
    # and the forcing dataset
    # and then drop the batch dimension
    training_xarray = (
        empty_dataset.combine_first(sfc_fields_xarray)
        .combine_first(pl_fields_xarray)
        .drop_vars(["batch"])
    )

    # Ensure we have the right order
    training_xarray = training_xarray.transpose(
        "batch",
        "time",
        "level",
        "lat",
        "lon",
    )

    # cfgrib sorts the pressure levels in descending order
    # we want them in ascending order

    training_xarray = training_xarray.reindex(
        level=sorted(training_xarray.level.values)
    )

    # And we want the grid south to north
    training_xarray = training_xarray.reindex(lat=sorted(training_xarray.lat.values))

    # # Shift times to timedelta 0
    training_xarray["time"] = [
        time_delta - time_deltas[0] for time_delta in time_deltas
    ]

    # Add datetimes as coordinates
    training_xarray.coords["datetime"].data[0, :] = datetimes

    if constants:
        # Add geopotential_at_surface and land_sea_mask back in
        x = xr.load_dataset(constants)

        for patch in ("geopotential_at_surface", "land_sea_mask"):
            LOG.info("PATCHING %s", patch)
            training_xarray[patch] = x[patch]

    return training_xarray, time_deltas



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

GRIB_TO_XARRAY_SFC = {
    "t2m": "2m_temperature",
    "msl": "mean_sea_level_pressure",
    "u10": "10m_u_component_of_wind",
    "v10": "10m_v_component_of_wind",
    "tp": "total_precipitation_6hr",
    "z": "geopotential_at_surface",
    "lsm": "land_sea_mask",
    "latitude": "lat",
    "longitude": "lon",
    "step": "batch",
    "valid_time": "datetime",
}

GRIB_TO_XARRAY_PL = {
    "t": "temperature",
    "z": "geopotential",
    "u": "u_component_of_wind",
    "v": "v_component_of_wind",
    "w": "vertical_velocity",
    "q": "specific_humidity",
    "isobaricInhPa": "level",
    "latitude": "lat",
    "longitude": "lon",
    "step": "batch",
    "valid_time": "datetime",
}


GRIB_TO_CF = {
    "2t": "t2m",
    "10u": "u10",
    "10v": "v10",
}



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import dataclasses
import datetime
import functools
import logging
import os
from functools import cached_property

import xarray
from ai_models.model import Model

from .input import create_training_xarray
from .output import save_output_xarray

LOG = logging.getLogger(__name__)


try:
    import haiku as hk
    import jax
    from graphcast import (
        autoregressive,
        casting,
        checkpoint,
        data_utils,
        graphcast,
        normalization,
    )
except ModuleNotFoundError as e:
    msg = "You need to install Graphcast from git to use this model. See README.md for details."
    LOG.error(msg)
    raise ModuleNotFoundError(e + msg)


class GraphcastModel(Model):
    download_url = "https://storage.googleapis.com/dm_graphcast/{file}"
    expver = "dmgc"

    # Download
    download_files = [
        (
            "params/GraphCast_operational - ERA5-HRES 1979-2021 - resolution 0.25 -"
            " pressure levels 13 - mesh 2to6 - precipitation output only.npz"
        ),
        "stats/diffs_stddev_by_level.nc",
        "stats/mean_by_level.nc",
        "stats/stddev_by_level.nc",
    ]

    # Input
    area = [90, 0, -90, 360]
    grid = [0.25, 0.25]

    param_sfc = [
        "lsm",
        "2t",
        "msl",
        "10u",
        "10v",
        "tp",
        "z",
    ]

    param_level_pl = (
        ["t", "z", "u", "v", "w", "q"],
        [50, 100, 150, 200, 250, 300, 400, 500, 600, 700, 850, 925, 1000],
    )

    forcing_variables = [
        "toa_incident_solar_radiation",
        # Not calling julian day and day here, due to differing assumptions with Deepmind
        # Those forcings are created by graphcast.data_utils
    ]

    use_an = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hour_steps = 6
        self.lagged = [-6, 0]
        self.params = None
        self.ordering = self.param_sfc + [
            f"{param}{level}"
            for param in self.param_level_pl[0]
            for level in self.param_level_pl[1]
        ]

    # Jax doesn't seem to like passing configs as args through the jit. Passing it
    # in via partial (instead of capture by closure) forces jax to invalidate the
    # jit cache if you change configs.
    def _with_configs(self, fn):
        return functools.partial(
            fn,
            model_config=self.model_config,
            task_config=self.task_config,
        )

    # Always pass params and state, so the usage below are simpler
    def _with_params(self, fn):
        return functools.partial(fn, params=self.params, state=self.state)

    # Deepmind models aren't stateful, so the state is always empty, so just return the
    # predictions. This is requiredy by the rollout code, and generally simpler.
    @staticmethod
    def _drop_state(fn):
        return lambda **kw: fn(**kw)[0]

    def load_model(self):
        with self.timer(f"Loading {self.download_files[0]}"):

            def get_path(filename):
                return os.path.join(self.assets, filename)

            diffs_stddev_by_level = xarray.load_dataset(
                get_path(self.download_files[1])
            ).compute()

            mean_by_level = xarray.load_dataset(
                get_path(self.download_files[2])
            ).compute()

            stddev_by_level = xarray.load_dataset(
                get_path(self.download_files[3])
            ).compute()

            def construct_wrapped_graphcast(model_config, task_config):
                """Constructs and wraps the GraphCast Predictor."""
                # Deeper one-step predictor.
                predictor = graphcast.GraphCast(model_config, task_config)

                # Modify inputs/outputs to `graphcast.GraphCast` to handle conversion to
                # from/to float32 to/from BFloat16.
                predictor = casting.Bfloat16Cast(predictor)

                # Modify inputs/outputs to `casting.Bfloat16Cast` so the casting to/from
                # BFloat16 happens after applying normalization to the inputs/targets.
                predictor = normalization.InputsAndResiduals(
                    predictor,
                    diffs_stddev_by_level=diffs_stddev_by_level,
                    mean_by_level=mean_by_level,
                    stddev_by_level=stddev_by_level,
                )

                # Wraps everything so the one-step model can produce trajectories.
                predictor = autoregressive.Predictor(
                    predictor,
                    gradient_checkpointing=True,
                )
                return predictor

            @hk.transform_with_state
            def run_forward(
                model_config,
                task_config,
                inputs,
                targets_template,
                forcings,
            ):
                predictor = construct_wrapped_graphcast(model_config, task_config)
                return predictor(
                    inputs,
                    targets_template=targets_template,
                    forcings=forcings,
                )

            with open(get_path(self.download_files[0]), "rb") as f:
                self.ckpt = checkpoint.load(f, graphcast.CheckPoint)
                self.params = self.ckpt.params
                self.state = {}

                self.model_config = self.ckpt.model_config
                self.task_config = self.ckpt.task_config

                LOG.info("Model description: %s", self.ckpt.description)
                LOG.info("Model license: %s", self.ckpt.license)

            jax.jit(self._with_configs(run_forward.init))
            self.model = self._drop_state(
                self._with_params(jax.jit(self._with_configs(run_forward.apply)))
            )

    @cached_property
    def start_date(self) -> "datetime":
        return self.all_fields.order_by(valid_datetime="descending")[0].datetime

    def run(self):
        with self.timer("Building model"):
            self.load_model()
        # all_fields = self.all_fields.to_xarray()

        with self.timer("Creating input data"):
            training_xarray, time_deltas = create_training_xarray(
                fields_sfc=self.fields_sfc,
                fields_pl=self.fields_pl,
                lagged=self.lagged,
                start_date=self.start_date,
                hour_steps=self.hour_steps,
                lead_time=self.lead_time,
                forcing_variables=self.forcing_variables,
                constants=self.override_constants,
            )

            if self.debug:
                training_xarray.to_netcdf("training_xarray.nc")

            input_xr, template, forcings = data_utils.extract_inputs_targets_forcings(
                training_xarray,
                target_lead_times=[
                    f"{int(delta.days * 24 + delta.seconds/3600):d}h"
                    for delta in time_deltas[len(self.lagged) :]
                ],
                **dataclasses.asdict(self.task_config),
            )

            if self.debug:
                input_xr.to_netcdf("input_xr.nc")
                forcings.to_netcdf("forcings_xr.nc")

        with self.timer("Doing full rollout prediction in JAX"):
            output = self.model(
                rng=jax.random.PRNGKey(0),
                inputs=input_xr,
                targets_template=template,
                forcings=forcings,
            )

            if self.debug:
                output.to_netcdf("output.nc")

        with self.timer("Saving output data"):
            save_output_xarray(
                output=output,
                write=self.write,
                target_variables=self.task_config.target_variables,
                all_fields=self.all_fields,
                ordering=self.ordering,
                lead_time=self.lead_time,
                hour_steps=self.hour_steps,
                lagged=self.lagged,
            )

    def patch_retrieve_request(self, r):
        if r.get("class", "od") != "od":
            return

        if r.get("type", "an") not in ("an", "fc"):
            return

        if r.get("stream", "oper") not in ("oper", "scda"):
            return

        if self.use_an:
            r["type"] = "an"
        else:
            r["type"] = "fc"

        time = r.get("time", 12)

        r["stream"] = {
            0: "oper",
            6: "scda",
            12: "oper",
            18: "scda",
        }[time]

    def parse_model_args(self, args):
        import argparse

        parser = argparse.ArgumentParser("ai-models graphcast")
        parser.add_argument("--use-an", action="store_true")
        parser.add_argument("--override-constants")
        return parser.parse_args(args)


model = GraphcastModel
